/*********************************************************
Chord keyboard scanner for MIDI Autoharp.

Written by Jim Arlow in 2018 for www.clearviewtraining.com
BSD license, all text above must be included in any redistribution.
**********************************************************/
#define DEBUG

#ifdef DEBUG
 #define DEBUG_PRINT(x)       Serial.print(x)
 #define DEBUG_PRINTDEC(x)    Serial.print(x, DEC)
 #define DEBUG_PRINTLN(x)     Serial.println(x)
 #define DEBUG_SERIAL_BEGIN() Serial.begin(9600)
#else
 #define DEBUG_PRINT(x)
 #define DEBUG_PRINTDEC(x)
 #define DEBUG_PRINTLN(x) 
 #define DEBUG_SERIAL_BEGIN()
#endif

// The last switch that was pressed
int CurrentSwitch = 10;

// Switches are arranged in NUM_BANKS banks of NUM_SWITCHES switches per bank
const int NUM_SWITCHES = 8;
const int SwitchPins[] = {9,8,7,6,5,4,3,2};

const int NUM_BANKS = 3;
const int BankPins[] = {12,11,10};

// 5 bit binary output of switch number, A0 = 2^0 to A5 = 2^4
const int NUM_OUT = 5;
const int OutPins[] = {A0, A1, A2, A3, A4};

// Data ready clock (positive edge)
const int ClockPin = 19;

// These are the switch numbers for the user defined switches
const int U0 = 7; 
const int U1 = 15; 
const int U2 = 23;

void setup() {
    //Serial.begin(9600); 
    DEBUG_SERIAL_BEGIN();
    DEBUG_PRINTLN("Chord Keyboard");
  
    // Setup the switch pins as inputs active LOW
    for (int i=0; i<NUM_SWITCHES;i++){
      pinMode(SwitchPins[i], INPUT_PULLUP);
      }
      
    // Setup the bank pins as outputs active LOW 
    for (int i=0; i<NUM_BANKS;i++){
      pinMode(BankPins[i], OUTPUT);
      setBank(i, HIGH);
      }
      
    // Setup the output pins
    for (int i=0; i<NUM_OUT;i++){
      pinMode(OutPins[i], OUTPUT);
      setOut(i, LOW);
      }
      
    // Set up the clock pin initialized to LOW
    pinMode(ClockPin, OUTPUT);
    setClock(LOW);
    
    // When the scanner powers up, no switches have been pressed,
    // so we need to set up a default output switch.
    CurrentSwitch=10;
    outputSwitchNumber();
}

// Test outputSwitchNumber(). To run:
// rename loop() to loopOLD() and rename this to loop()
void TESTloop() {
  for (int i =0; i < 24; i++)
  {
    CurrentSwitch=i;
    outputSwitchNumber();
    delay(1000);
  }
}

void loop() {
  
  // Scan banks
  for (int i = 0; i < NUM_BANKS; i++){
    
    // Select the ith bank (active LOW)
    setBank(i, LOW);
    
    // Scan the switches in the bank
    for (int j = 0; j < NUM_SWITCHES; j++){
      if (getSwitch(j) == LOW){
      updateCurrentSwitch(i, j);
      }
    }
    
    // Deselect the ith bank
    setBank(i, HIGH);    
  }

}


void setClock(int state){
  digitalWrite(ClockPin, state);
}


void setOut(int i, int state){ 
  digitalWrite(OutPins[i], state);
}

void setBank(int i, int state){ 
  digitalWrite(BankPins[i], state);
}

int getSwitch(int i){
  return digitalRead(SwitchPins[i]);
}

// Given the bank, b and the switch s, calculate the key number and
// output it
void updateCurrentSwitch(int b, int s){
  int newSwitch = b*NUM_SWITCHES + s;
  
  // The user defined switches should NOT latch
  if ((newSwitch == U0) || (newSwitch == U1) || (newSwitch == U2)){
    CurrentSwitch = newSwitch;
    outputSwitchNumber();
    // Delay for 0.5 seconds to account for any switch bounce
    delay(500);
    return;
  }
  
  // Normal chord switches latch i.e. the output only changes when the chord changes
  if (newSwitch != CurrentSwitch)
  {
    CurrentSwitch = newSwitch;
    DEBUG_PRINTLN("NEW SWITCH = " + String(CurrentSwitch));
    outputSwitchNumber();
  }
}

// Output the switch number as a binary number along with
// a clock. Valid switch number is on the POSITIVE EDGE
void outputSwitchNumber(){
  // The clock is a positive edge trigger, so set to LOW
  setClock(LOW);
    
  for (int i=0; i<NUM_OUT;i++){
    setOut(i, bitRead(CurrentSwitch, i));
  }
  
  // The clock is a positive edge trigger, set to HIGH and create the edge
  setClock(HIGH);
}

