/*********************************************************
This is a library for the MPR121 12-channel Capacitive touch sensor

Designed specifically to work with the MPR121 Breakout in the Adafruit shop 
  ----> https://www.adafruit.com/products/

These sensors use I2C communicate, at least 2 pins are required 
to interface

Adafruit invests time and resources providing this open source code, 
please support Adafruit and open-source hardware by purchasing 
products from Adafruit!

Written by Limor Fried/Ladyada for Adafruit Industries.  
BSD license, all text above must be included in any redistribution
**********************************************************/

/*********************************************************
This code is derived by Jim Arlow from MPR121test.ino by Limor Fried/Ladyada.

Designed for use with an Arduino Uno connected to an 
Adafruit MPR121 12-Key Capacitive Touch Sensor Breakout.
Adafruit makes great stuff! Support them!

Written by Jim Arlow in 2018 for www.clearviewtraining.com
BSD license, all text above must be included in any redistribution.
**********************************************************/

// Uncomment to put the instrument in debug mode
//#define DEBUG

#ifdef DEBUG
 #define DEBUG_PRINT(x)       Serial.print(x)
 #define DEBUG_PRINTDEC(x)    Serial.print(x, DEC)
 #define DEBUG_PRINTLN(x)     Serial.println(x)
 #define SERIAL               9600                // For debugging
#else
 #define DEBUG_PRINT(x)
 #define DEBUG_PRINTDEC(x)
 #define DEBUG_PRINTLN(x) 
 #define SERIAL               31250               // For MIDI
#endif

#include <Wire.h>
#include "Adafruit_MPR121.h"

// You can have up to 4 on one i2c bus but one is enough for testing!
Adafruit_MPR121 cap = Adafruit_MPR121();

// Keeps track of the last strings touched so we know when strings are 'released'
uint16_t lasttouched = 0;
uint16_t currtouched = 0;

// Number of input pins from chord keyboard
const int NUM_IN = 5;
// 5 bit binary input of switch number from Chord Keyboard. A0 = 2^0 to A5 = 2^4
const int InPins[] = {7, 6, 5, 4, 3};


////////////////////////////////////////////////////////////////////
// Interrupt pins
////////////////////////////////////////////////////////////////////
// This pin recieves an interrrupt from the Chord Keyboard that indicates a new chord
const int INTERRUPT_PIN = 2;        // Octave down button on MIDI Lyre shield
// This pin causes a MIDI panic and sends a note off to all 127 notes
const int PANIC_INTERRUPT_PIN = 3;  // Octave up button on MIDI Lyre shield
////////////////////////////////////////////////////////////////////

// Chord patterns
const int Maj[] = {0, 4, 7, 12, 16, 19, 24, 28, 31, 36, 40,  43};
const int Sev[] = {0, 4, 7, 10, 12, 16, 19, 22, 24, 28, 31, 34};
const int Min[] = {0, 3, 7, 12, 15, 19, 24, 27, 31, 36, 39, 43};

// Pitch classes
const int C   = 0; 
const int Cs  = 1; const int Db   = 1; 
const int D   = 2; 
const int Ds  = 3; const int Eb   = 3; 
const int E   = 4;
const int F   = 5; 
const int Fs  = 6; const int Gb   = 6;
const int G   = 7; 
const int Gs  = 8; const int Ab   = 8;
const int A   = 9;
const int As  = 10; const int Bb  = 10;
const int B   = 11;

// These are the switch numbers for the user defined switches
const int U0 = 7;   // Octave up
const int U1 = 15;  // Octave down
const int U2 = 23;  // Lyre mode

// These are the Tonics (plus user defined) for the Americana layout
const int Tonics[] = {
  Eb,F,C,G,D,A,E,U0,
  Bb,F,C,G,D,A,E,U1,
  G,D,A,E,B,Fs,B,U2
};

// Useful for debugging
const String TonicsStr[] = {
  "Eb","F","C","G","D","A","E","U0",
  "Bb","F","C","G","D","A","E","U1",
  "G","D","A","E","B","F#","B","U2"
};

// These are the chords for the Americana layout
const int Chords[24] = {
  Maj,Sev,Sev,Sev,Sev,Sev,Sev,Maj,
  Maj,Maj,Maj,Maj,Maj,Maj,Maj,Maj,
  Min,Min,Min,Min,Min,Min,Sev,Maj,
};

const String ChordsStr[24] = {
  "","7","7","7","7","7","7","",
  "","","","","","","","",
  "m","m","m","m","m","m","7","",
};

// The octave range
const int MAX_OCTAVE = 6;
const int MIN_OCTAVE = 0;

////////////////////////////////////////////////////////////////////
// Variables that control the tuning of the strings               //
////////////////////////////////////////////////////////////////////
volatile int Strings[] = {0, 4, 7, 10, 12, 16, 19, 22, 24, 28, 31, 34};     // The current string tuning
volatile int Octave = 4;                                                    // The current Octave. Middle C = MIDI 60
volatile int CurrentKey = 10;                                               // The current key that was pressed
const int LyreTuning[] =  {-5, -3, -1, 0, 2, 4, 5, 7, 9, 11, 12, 14};       // C Major Mixolydian mode for Lyre
////////////////////////////////////////////////////////////////////

// Flash the internal LED
void flashLED(int d=100){
    digitalWrite(LED_BUILTIN, HIGH);    // turn the LED on (HIGH is the voltage level)
    delay(d);                           // wait for d/1000 seconds
    digitalWrite(LED_BUILTIN, LOW);     // turn the LED off by making the voltage LOW
    delay(d);                           // wait for d/1000 seconds
}


// Plays a MIDI note on channel 1 with default velocity of 127
void noteOn(int note, int velocity=127){
  Serial.write(0x90);
  Serial.write(note);
  Serial.write(velocity); 
}

// Turns the given MIDI note off
void noteOff(int note){
  Serial.write(0x80);
  Serial.write(note);
  Serial.write(0);
}

void setup() { 
  // Open the serial port for MIDI or debugging
  Serial.begin(SERIAL);
    
  delay(1000);
  DEBUG_PRINTLN("Chord keyboard test");
    
  // Initialize digital pin LED_BUILTIN as an output.
  pinMode(LED_BUILTIN, OUTPUT); 
  
  // Setup the input pins to receive the key number from the Chord Keyboard
  for (int i=0; i<NUM_IN;i++){
      pinMode(InPins[i], INPUT);
    }  
      
  // Set up the interrupt from chord keyboard
  pinMode(INTERRUPT_PIN, INPUT_PULLUP);      
  attachInterrupt(digitalPinToInterrupt(INTERRUPT_PIN), interruptHandler, RISING);

  // Set up the MIDI panic interrupt
  pinMode(PANIC_INTERRUPT_PIN, INPUT_PULLUP);      
  attachInterrupt(digitalPinToInterrupt(PANIC_INTERRUPT_PIN), panicInterruptHandler, FALLING);

    
  // Flash continuously to indicate an error
  while (!cap.begin(0x5A)) { flashLED(); } 
  
  // Flash 20 times to indicate success and to give the sensor time to stabilize
  for (int i=0;i<20;i++){ flashLED(); } 
  
  // Set the initial tuning of the strings
  int* newStrings = Chords[CurrentKey];
  for (int i = 0; i < 12; i++){
    Strings[i] = newStrings[i] + Tonics[CurrentKey] + Octave*12;
  }

}


// Detect a "pluck", which is a touch then release.
void negativeEdge(){
  // Get the currently touched strings
  currtouched = cap.touched();
  
  for (uint8_t i=0; i<12; i++) {
    // if string i was touched and is now released, play the  note
    if (!(currtouched & _BV(i)) && (lasttouched & _BV(i)) ) { 
      noteOn(Strings[i]); 
    }
  }

  // Reset our state
  lasttouched = currtouched;  
}

// Detect a "press", which is a touch (NOT USED).
void positiveEdge(){
  // Get the currently touched strings
  currtouched = cap.touched();
  
  for (uint8_t i=0; i<12; i++) {
    // if string i was not touched and is now touched, damp the note
    if ((currtouched & _BV(i)) && !(lasttouched & _BV(i)) )  { 
      noteOff(Strings[i]); 
    }
  }

  // Reset our state
  lasttouched = currtouched;  
}


// Handle the interrupt from the Chord Keyboard
void interruptHandler(){
  // We will turn off interrupts to be sure that this executes atomically
  noInterrupts();
  
  // Read the key number from the chord keyboard  
  volatile int key = 0;
  bitWrite(key, 0, digitalRead(InPins[0]));
  bitWrite(key, 1, digitalRead(InPins[1]));
  bitWrite(key, 2, digitalRead(InPins[2]));
  bitWrite(key, 3, digitalRead(InPins[3]));
  bitWrite(key, 4, digitalRead(InPins[4]));
  
  // Handle the key number. Only change the CurrentKey if the key corresponds to a note
  switch(key){
    case U0:  // Increment Octave
      Octave = Octave + 1;
      if (Octave > 6) {Octave = 6;}
      DEBUG_PRINT("U0 ");
      DEBUG_PRINTDEC(Octave);
      DEBUG_PRINTLN("");
      break;
    case U1:  // Decrement Octave
      Octave = Octave - 1;
      if (Octave < 0) {Octave = 0;}
      DEBUG_PRINT("U1 ");
      DEBUG_PRINTDEC(Octave);
      DEBUG_PRINTLN("");
      break;
    case U2:  // Switch to Lyre mode
      //Octave = 5; // Set the Lyre to middle C
      for (int i = 0; i < 12; i++){
        Strings[i] = LyreTuning[i] + Tonics[CurrentKey] + Octave*12;
      }  
      DEBUG_PRINT("U2 ");
      DEBUG_PRINTDEC(Octave);
      DEBUG_PRINTLN("");
      return;
    default:
      // The key corresponds to a note, so set CurrentKey to the value of key
      CurrentKey = key;
      printInputs();
  }
  
  // Tune the strings
  int* newStrings = Chords[CurrentKey];
  for (int i = 0; i < 12; i++){
    Strings[i] = newStrings[i] + Tonics[CurrentKey] + Octave*12;
  }
  
  interrupts();
}

// If there is a MIDI mess up, this panic mode sends a MIDI note off to all notes.
// It is assigned to the octave down switch on the MIDI Lyre shield.
void panicInterruptHandler(){
  // We will turn off interrupts to be sure that this executes atomically
  noInterrupts();

  for(int i=0; i<128; i++){
    noteOff(i);
  }
  
  interrupts();
}

// For debugging
void printInputs(){
  // Most significant bit to least significant bit
  DEBUG_PRINT(digitalRead(InPins[4]));
  DEBUG_PRINT(digitalRead(InPins[3]));
  DEBUG_PRINT(digitalRead(InPins[2]));
  DEBUG_PRINT(digitalRead(InPins[1]));
  DEBUG_PRINT(digitalRead(InPins[0]));   
  DEBUG_PRINT(" ");
  
  // Decimal
  DEBUG_PRINT(CurrentKey);
  DEBUG_PRINT(" ");
  
  // Chord
  DEBUG_PRINT(TonicsStr[CurrentKey]);
  DEBUG_PRINTLN(ChordsStr[CurrentKey]);
}

void loop() {
  negativeEdge();
}



